/*
 * $Id: morphy-usbio.c,v 1.5 2000/12/26 15:38:14 nari Exp $
 * Driver for Morphy USB IO
 *
 *    http://www.morphoyplanning.co.jp/
 *
 */
#define  _LOOSE_KERNEL_NAMES
#include <linux/module.h>
#include <linux/socket.h>
#include <linux/miscdevice.h>
#include <linux/list.h>
#include <linux/types.h>
#include <linux/vmalloc.h> 
#include <linux/slab.h> 
#include <linux/init.h>
#include <asm/uaccess.h> 
#include <asm/atomic.h>
#include <linux/delay.h>
#include <linux/usb.h>
#include <linux/ctype.h>
#include <asm/segment.h>

#include "morphy-usbio.h"

typedef struct musbio
{
  struct usb_device *dev;
  unsigned int ifnum;
  unsigned char buf[2];
  unsigned int mode;
  struct list_head file_list;

  urb_t        *urb;
  int        bReading;
  int        useCount;
  struct wait_queue *wait_reply;
  struct wait_queue *wait_busy;
  
  unsigned char iBuf[8];
} musbio_t, *pmusbio_t;

typedef struct musbio_file
{
  struct list_head list_head;
  struct musbio *musbio;
  unsigned int mode;
  unsigned int port;
} musbio_file_t;

void musbio_read_data(musbio_file_t *mf);

#ifdef DEBUG
#undef dbg
#define dbg(format, arg...) printk(KERN_NOTICE __FILE__ ": " format "\n" , ## arg)
#endif

static musbio_t *musbio_devices[NRMUSBIO];

static int musbio_open(struct inode *inode, struct file *filp)
{
  int minor  = MINOR(inode->i_rdev);
  int devnum = (minor - MUSBIO_MINOR)/ 2;
  int port   = minor % 2;
  musbio_t *s;
  musbio_file_t *mf;

  if( devnum < 0 || devnum >= NRMUSBIO)
    return -EIO;
  s = musbio_devices[devnum];
  if(!s) return -EIO;
  mf = kmalloc(sizeof(musbio_file_t), GFP_KERNEL);
  if(!mf) return -ENOMEM;
  memset(mf, 0, sizeof(musbio_file_t));
  list_add(&mf->list_head, &s->file_list);
  mf->musbio = s;
  mf->mode = s->mode;
  mf->port = port;
  filp->private_data = mf;
  return 0;
}

static int musbio_release (struct inode *inode, struct file *file)
{
  musbio_file_t *mf = (musbio_file_t *)file->private_data;
  list_del(&mf->list_head);
  kfree(mf);
  return 0;
}

static int musbio_write(struct file *filp, const char *buf,
			size_t count, loff_t *off)
{
  musbio_file_t *mf = (musbio_file_t *)filp->private_data;
  musbio_t *s = mf->musbio;
  int data,data_valid;
  const char *p, *p_end;
  char c;
  
  if(!s) return -ENODEV;
  
  p = buf;
  p_end = buf + count;

  while(p < p_end){
    data = data_valid = 0;
    if(mf->mode == MUSBIO_MODE_BINTEXT){
      for(get_user(c,p); isspace(c) && p < p_end; get_user(c,++p));
      for(get_user(c,p);!isspace(c) && p < p_end; get_user(c,++p)){
	data <<= 1;
	if(c != '0') data ++;
	data_valid = 1;
      }
    }
    else if(mf->mode == MUSBIO_MODE_HEXTEXT){
      for(get_user(c,p); isspace(c) && p < p_end; get_user(c,++p));
      for(get_user(c,p);!isspace(c) && p < p_end; get_user(c,++p)){
	data <<= 4;
	if(isdigit(*p)) data += *p - '0';
	else if(isxdigit(c)){
	  if(isupper(c))
	    data += c - 'A' + 10;
	  else
	    data += c - 'a' + 10;
	}
	data_valid = 1;
      }
    }
    else { /* binary */
      get_user(data,p++);
      if(mf->port == MUSBIO_PORT01 && p < p_end){
	get_user(c,p++);
	data =  (data & 0xff) |(c  << 8);
      }
      data_valid = 1;
    }

    if(data_valid){
      if(mf->port == MUSBIO_PORT0){
	s->buf[0] = 1;
	s->buf[1] = data;
	usb_set_report(s->dev, s->ifnum, 2, 0, s->buf, 2);
      }
      else if(mf->port == MUSBIO_PORT1){
	s->buf[0] = 2;
	s->buf[1] = data;
	usb_set_report(s->dev, s->ifnum, 0, 2, s->buf, 2);
      }
      else if(mf->port == MUSBIO_PORT01){
	s->buf[0] = 1;
	s->buf[1] = data;
	usb_set_report(s->dev, s->ifnum, 0, 2, s->buf, 2);
	s->buf[0] = 2;
	s->buf[1] = (data >> 8);
	usb_set_report(s->dev, s->ifnum, 0, 2, s->buf, 2);
      }
      else {
	err("write unknown port=%d", mf->port);
      }
    }
  }
  return count;
}

static ssize_t musbio_read(struct file *filp, char *buf,
		       size_t buflen, loff_t *ppos)
{
  musbio_file_t *mf = (musbio_file_t *)filp->private_data;
  musbio_t *s = mf->musbio;
  int len,data = 0;
  char *q;
  char lbuf[20];

  if (*ppos)
    return -ESPIPE;
  if(!s) return -ENODEV;
  if(buflen < 20) return -EINVAL;

  while(s->useCount){
    if(filp->f_flags & O_NONBLOCK) return -EAGAIN;
    dbg("wait busy");
    interruptible_sleep_on(&s->wait_busy);
  }
  s->useCount++;
  
  if(mf->port == MUSBIO_PORT0){
    s->buf[0] = 3;
    usb_set_report(s->dev, s->ifnum, 2, 0, s->buf, 1);
    musbio_read_data(mf);
    data = s->iBuf[1];
  }
  else if(mf->port == MUSBIO_PORT1){
    s->buf[0] = 4;
    usb_set_report(s->dev, s->ifnum, 2, 0, s->buf, 1);
    musbio_read_data(mf);
    data = s->iBuf[1];
  }
  else if(mf->port == MUSBIO_PORT01){
    s->buf[0] = 3;
    usb_set_report(s->dev, s->ifnum, 0, 2, s->buf, 1);
    musbio_read_data(mf);
    data = s->iBuf[1];
    s->buf[0] = 4;
    usb_set_report(s->dev, s->ifnum, 0, 2, s->buf, 1);
    musbio_read_data(mf);
    data += s->iBuf[1] << 8;
  }
  
  len = 0;
  if(mf->mode == MUSBIO_MODE_BINTEXT){
    int mask;
    int i,width;
    if(mf->port == MUSBIO_PORT0)  width = 8;
    else if(mf->port == MUSBIO_PORT1)  width = 4;
    else if(mf->port == MUSBIO_PORT01) width = 12;
    else width = 16;
    
    q = lbuf;
    mask = 1 << (width - 1);
    for(i=0; i<width;i++,mask >>= 1){
      *q++ = (data & mask) ? '1' : '0';
    }
    *q++ = '\n';
    len = q - lbuf;
  }
  else if(mf->mode == MUSBIO_MODE_HEXTEXT){
    sprintf(lbuf,"%04x\n", data);
    len = strlen(lbuf);
  }
  else {
    if(mf->port == MUSBIO_PORT01) {
      lbuf[0] = data & 0xff;
      lbuf[1] = data >> 8;
      len = 2;
    }
    else {
      lbuf[0] = data;
      len = 1;
    }
  }
  copy_to_user(buf, lbuf, len);
  if(--s->useCount ==0)
    wake_up_interruptible(&s->wait_busy);
  return len;
}

static int musbio_ioctl(struct inode *inode, struct file *filp,
			unsigned int cmd, unsigned long arg)
{
  musbio_file_t *mf = (musbio_file_t *)filp->private_data;
  musbio_t *s = mf->musbio;
  int pulse_bit;
  int value;

  if(!s) return -ENODEV;
  if(_IOC_TYPE(cmd) != MUSBIO_IOC_MAGIC) return -EINVAL;
  if(_IOC_NR(cmd) > MUSBIO_IOC_MAXNR)    return -EINVAL;

  switch(cmd){
  case MUSBIO_IOCTDEFMODE:
    s->mode = arg;
    break;
    
  case MUSBIO_IOCQDEFMODE:
    return s->mode;
    
  case MUSBIO_IOCTMODE:
    mf->mode = arg;
    break;
    
  case MUSBIO_IOCQMODE:
    return mf->mode;
    
  case MUSBIO_IOCTPORT:
    mf->port = arg;
    break;
    
  case MUSBIO_IOCQPORT:
    return mf->port;
    
  case MUSBIO_IOCQVER:
    return s->dev->descriptor.idProduct;
    
  case MUSBIO_IOCTW0P1:          /* write P0, pulse P1 */
    if(s->dev->descriptor.idProduct < 0x1003)
      return -EINVAL;
    pulse_bit = arg >> 8;
    value     = arg & 0xff;
    if(pulse_bit > 3) return -EINVAL;
    
    s->buf[0] = 0x10 + pulse_bit;
    s->buf[1] = value;
    usb_set_report(s->dev, s->ifnum, 0, 2, s->buf, 2);
    break;
    
  case MUSBIO_IOCQP1R0:          /* pulse P1 ,read P0 */
    /* not implemented yet */
    return -EINVAL;
    
  default:
    return -EINVAL;
  }
  return 0;
}

static struct file_operations musbio_fops =
{
  open:		musbio_open,
  release:      musbio_release,
  write:        musbio_write,
  read:         musbio_read,
  ioctl:        musbio_ioctl,
};

void musbio_int_complete(purb_t purb);

static void *musbio_probe(struct usb_device *dev, unsigned int ifnum)
{
  musbio_t *s,**ps;
  struct usb_interface_descriptor *interface;
  int i;
  
  dbg("musbio_probe(%04x:%04x) ifnum=%u",dev->descriptor.idVendor,
      dev->descriptor.idProduct, ifnum);

  if(dev->descriptor.bNumConfigurations != 1) return NULL;
  interface = dev->config[0].interface[ifnum].altsetting + 0;
  
  if(dev->descriptor.idVendor==0x0bfe &&
     dev->descriptor.idProduct>=0x1000 &&
     dev->descriptor.idProduct<=0x10ff){
    dbg("mophy-usbio found.");

    for(i=0;i<NRMUSBIO;i++)
      if(musbio_devices[i] == NULL)
	goto found;
    info("device full");
    return NULL;
    
  found:
    info("dev %d assigned.",i);
    MOD_INC_USE_COUNT;
    ps = &musbio_devices[i];
    *ps = s = kmalloc(sizeof(musbio_t), GFP_KERNEL);
    if(!s) return NULL;
    memset(s, 0, sizeof(musbio_t));
    s->dev = dev;
    s->ifnum = ifnum;
    s->urb = usb_alloc_urb(0);
    FILL_INT_URB(s->urb, dev, usb_rcvintpipe(dev, 1),
		 s->iBuf, sizeof(s->iBuf),
		 musbio_int_complete, s, 4);
    INIT_LIST_HEAD(&s->file_list);

    s->urb->start_frame = 0;

    return ps;
  }
  return NULL;
}

static void musbio_disconnect (struct usb_device *usbdev, void *ptr)
{
  musbio_t         **ps = (musbio_t **)ptr,*s=*ps;
  musbio_file_t    *mf;
  struct list_head *pos;
  
  dbg("musbio_disconnect");

  if(s->bReading)
    usb_unlink_urb(s->urb);
  
  for (pos = s->file_list.next; pos != &(s->file_list); pos = pos->next){
    mf = (musbio_file_t *)pos;
    mf->musbio = NULL;
  }
  MOD_DEC_USE_COUNT;
  kfree(*ps);
  *ps = NULL;
}

static struct usb_driver musbio_driver =
{
  "morphy-usbio",
  musbio_probe,
  musbio_disconnect,
  {NULL, NULL},
  &musbio_fops, 
  MUSBIO_MINOR
};

#ifdef MODULE
MODULE_AUTHOR ("Hiroshi Narimatsu , nari@mxb.mesh.ne.jp");
MODULE_DESCRIPTION ("Morphy USB-IO Driver for Linux (c)2000");

int init_module (void)
{
  dbg("morpy-usbio: init_module");
  if(usb_register(&musbio_driver))
    return -1;
  return 0;
}

void cleanup_module (void)
{
  dbg("morpy-usbio: cleanup_module");
  usb_deregister(&musbio_driver);
}

#endif

void musbio_int_complete(purb_t purb)
{
  musbio_t *s = (musbio_t *)purb->context;
  wake_up_interruptible(&s->wait_reply);
}

void musbio_read_data(musbio_file_t *mf)
{
  musbio_t *s = mf->musbio;
  
  usb_submit_urb(s->urb);
  s->bReading = 1;
  interruptible_sleep_on(&s->wait_reply);
  usb_unlink_urb(s->urb);
}
